/* N-body calculation using Greenspan 1st-order discrete mechanics.    */
/* Incorporates xyz-parallelism, uses domain pointers, combines        */
/* particle and pair domains, diagonal processors store particles      */
/* Spreading coded in C*, reduced overhead in accumulation,            */
/*  more globals, Booleans stored in char fields.                      */

#include <stdio.hs>
#include <math.hs>
#include "cmtime.hs"

/* Max number of bodies in the calculation.  N = 64 fits in 16k procs
   N must be a power of 2, 4N^2 <= no. of v. procs. */
#define N 64

int NB;      /* Number of bodies in system, <= N */
int LNB;     /* Largest power of 2 < NB */
float G;     /* Gravitational constant */

/* Particles, and pairs of particles.  Diagonal entries have particle	*/
/* states.  Some variables could be overloaded. 			*/
/* act  - processor is active pair					*/
/* sm   - processor participates in acceleration sum			*/
/* here - processor represents particle					*/
/* sp   - processor participates in spreading                           */
/* smk  - control for spreading, accumulation, when procs change        */
/* th   - 'this', for convenience                                       */

domain pair { int i, j, smk, th;  char act, sm, here, sp;
	      float x0, x1, v0, v1, m, d0, r0, ac, w;
	      domain pair *pi, *pj;  domain pair *r; };

domain pair P[3][N][N];

/* Pairs will be represented symmetrically.  The function setup
   (below) initializes these so that pi, pj point to appropriate
   particle in corresponding coordinate.   */

void pair::setup()
 { int offset, c;  char smm;  mono int k;
   for ( LNB = 1; LNB < NB; LNB <<= 1 ) ;
   sp = (i<LNB) && (j < LNB);  LNB >>= 1;
   th = (int) this;
   offset = (this - &P[0][0][0]);
   c = offset/(N*N);  i = (offset/N)%N;  j = offset%N;
   sm = (i < NB) && (j < NB);
   act = sm && (i != j);  here = sm && (i == j);
   pi = &P[c][i][i];  pj = &P[c][j][j];
   r = (domain pair *)( (c<2) ? th+N*N : th-2*N*N );
   smm = 1;  smk = i^j;  /* xor */
   for ( k = 1; k < NB; k <<= 1 )
    if (smm && (k & smk)) { smk = k;  smm = 0; }
 }

FILE *datap;      /* Stream for reading data */

/* Load inital positions, velocities, also masses, gravitational const. */

void init()
 { int k;  float a, b, c;
   [domain pair].
   {for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e%e%e", &a, &b, &c);
       P[0][k][k].x0 = a;  P[1][k][k].x0 = b;  P[2][k][k].x0 = c;
     };
    for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e%e%e", &a, &b, &c);
       P[0][k][k].v0 = a;  P[1][k][k].v0 = b;  P[2][k][k].v0 = c;
     };
    for ( k = 0; k < NB; k++ )
      { fscanf(datap, "%e", &a);
        P[0][k][k].m = P[1][k][k].m = P[2][k][k].m = a;
      };
    fscanf(datap, "%e", &G);
   }
 }

void pair::print_state()
 { mono int k;
   printf(
  "\n       x         y         z         vx        vy        vz        m\n");
   for ( k = 0; k < NB; k++ )
     printf("%2d %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.2e\n",
            k, P[0][k][k].x0, P[1][k][k].x0, P[2][k][k].x0,
               P[0][k][k].v0, P[1][k][k].v0, P[2][k][k].v0, P[0][k][k].m);
 }

/* smk[i,j] is a power of 2 corresponding to the stage at which proc[i,j]
   sends it's acceleration to partner and drops out of accumulation. It
   also corresponds to the stage at which proc[i,j] begins to send values
   during spreading operations.                                          */

/* Spread xv from diagonal along i-axis.  Returns result in w.  */
/* May spread to power of 2 just larger than NB                 */

void pair::spreadi(xv)
  float xv;
  { char smm;  mono int k, b;
    smm = here;  
    for ( k = 2*LNB, b = LNB*N;  k > 1;  k >>= 1, b >>= 1 )
     if ( smm || ( k == smk) )
      { smm = 1;  ((domain pair *)(th^b))->xv = xv; }
    w = xv;
  }

/* Spread xv from diagonal along j-axis.  Returns result in w.  */
/* May spread to power of 2 just larger than NB                 */

void pair::spreadj(xv)
  float xv;
  { char smm;  mono int k;
    smm = here;  
    for ( k = 2*LNB; k > 1; )
     if ( smm || (k == smk) )
      { smm = 1;  k >>= 1;  ((domain pair *)(th^k))->xv = xv; }
    w = xv;
  }

/* Sum accelerations - call with sp's active */

void pair::sumac()
 { char smm;  mono int k;
      smm = 1;
      for ( k = 1; k < NB; k <<= 1 )
	{ if ( smm && (k == smk) )
	     { ((domain pair *)(th^k))->w = ac;  smm = 0; }
	  else ac += w; }
 } 

/* Calculate accelerations - called with sp's active */

void pair::accel()
 {    float d1, r1;  char smm;
      mono int k;
      ac = 0.0;  spreadi(x1);  d1 = w;  spreadj(x1);  x1 = w;
      if (act)
      { d1 = d1 - x1;  r1 = d1*d1;  r->ac = r1;
       ac += r1;  r->w = ac;  r1 = sqrt(w + r1);
       ac = G*m*(d0+d1)/(r0*r1*(r0+r1)); };
      sumac();
 }

/* Calculate initial accelerations - called with sp's active */

void pair::accel1()
 {    char smm;  mono int k;
      ac = 0.0;  spreadi(x0);  d0 = w;  spreadj(x0);  x0 = w;
      if (act)
      { d0 = d0 - x0;  r0 = d0*d0;  r->ac = r0;
        ac += r0;  r->w = ac;  w += r0;  r0 = sqrt(w);
        ac = G*m*r0/(r0*w);  };
      sumac();
 }


/* Greenspan's 1st-order method.  Performs ns steps of time dt
   with solution tolerance eps.  Returns number of evaluations.
   Leaves number of iterations which were stopped at limit in cfail.  */

int cfail;   /* Counts number of iterations that failed to converge */

int green1(ns, dt, eps)
  int ns;  float dt, eps;
 { int k, ic, v, lim=10;  float hdt, mx, my;
   v = 0;  hdt = 0.5*dt;
   [domain pair]. if (sp)  /* Everything in "spreading" procs */
   { float nx;
     spreadi(m);  m = w;
     for (k = 0; k < ns; k++)
       { ic = 0;  accel1();
	 /* This is a do - while - repeat loop: */
   	 do
          { v++;  ic++;
	    if (here)
	    { nx = v0 + dt*ac;  mx = (>?= fabs(nx - v1));
	      v1 = nx;
	      nx = x0 + hdt*(v0 + v1);
	      my = (>?= fabs(nx - x1));
	      if (my > mx) mx = my;
	      x1 = nx; }
	      if ( (mx < eps) || ( ic == lim ) ) break;
	      accel(); }
	 while (1);
         if (here) { x0 = x1;  v0 = v1; };  if (ic == lim) cfail++;
       } ;   /* end for */
    }  /* end  [domain pair] */
   return(v);
 }


void main()
 { int ns, np, nl, nv, k, vl;  float dt, eps;
   char fnm[50];
     /* Performs np integrations, each ns steps of dt */
     printf("Data file> ");  scanf("%s", fnm);
     datap = fopen(fnm, "r");
     fscanf(datap, "%d", &NB);
     [domain pair].{ setup(); };
     init();
     fscanf(datap, "%d%d%e%e", &ns, &np, &dt, &eps);
     fclose(datap);
     printf("\n   G = %13.6e  delta-t = %12.5e  epsilon = %12.5e\n",
                   G, dt, eps);
     printf("\n simulated time = 0");
     [domain pair].{ print_state(); };
     for ( k = 0; k < np; k++ )
      { CM_start_timer(1);
        cfail = 0;
        vl = green1(ns, dt, eps);
        CM_stop_timer(1);  CM_reset_timer();
        printf(
	 "\n simulated time = %9.3f  %d evaluations  %d iterations limited\n",
	  (k + 1)*ns*dt, vl, cfail);
        [domain pair].{ print_state(); };
      };
 }

